#include "engine/db/obj_prototypes.h"
#include "engine/entities/char_data.h"
#include "engine/core/handler.h"
#include "gameplay/clans/house.h"
#include "gameplay/mechanics/liquid.h"
#include "gameplay/fight/pk.h"
#include "engine/core/utils_char_obj.inl"

//    .  ,    .
const int kMaxLabelLength = 32;

void DoSign(CharData *ch, char *argument, int/* cmd*/, int/* subcmd*/) {
	char arg1[kMaxInputLength];
	char arg2[kMaxInputLength];

	char arg3[kMaxInputLength];
	char arg4[kMaxInputLength];

	ObjData *target = nullptr;
	int erase_only = 0; // 0 --   , 1 --  
	int clan = 0; //  ,  . ,  
	int no_target = 0; //    ,   

	char *objname = nullptr;
	char *labels = nullptr;

	if (ch->IsNpc())
		return;

	half_chop(argument, arg1, arg2);

	if (!strlen(arg1))
		no_target = 1;
	else {
		if (!strcmp(arg1, "")) { //   arg1 "",   arg2     
			clan = 1;
			if (strlen(arg2)) {
				half_chop(arg2, arg3, arg4); //  arg3   
				objname = str_dup(arg3);
				if (strlen(arg4))
					labels = str_dup(arg4);
				else
					erase_only = 1;
			} else {
				no_target = 1;
			}
		} else { //  ""  , ,   arg1       arg2
			if (strlen(arg1)) {
				objname = str_dup(arg1);
				if (strlen(arg2))
					labels = str_dup(arg2);
				else
					erase_only = 1;
			} else {
				no_target = 1;
			}
		}
	}

	if (no_target) {
		SendMsgToChar("  ?\r\n", ch);
	} else {
		if (!(target = get_obj_in_list_vis(ch, objname, ch->carrying))) {
			sprintf(buf, "   \'%s\'.\r\n", objname);
			SendMsgToChar(buf, ch);
		} else {
			if (erase_only) {
				target->remove_custom_label();
				act("    $o5.", false, ch, target, nullptr, kToChar);
			} else if (labels) {
				if (strlen(labels) > kMaxLabelLength)
					labels[kMaxLabelLength] = '\0';

				//  
				for (int i = 0; labels[i] != '\0'; i++)
					if (labels[i] == '~')
						labels[i] = '-';

				std::shared_ptr<custom_label> label(new custom_label());
				label->text_label = str_dup(labels);
				label->author = ch->get_uid();
				label->author_mail = str_dup(GET_EMAIL(ch));

				const char *msg = "  $o3 ,      .";
				if (clan && ch->player_specials->clan) {
					label->clan_abbrev = str_dup(ch->player_specials->clan->GetAbbrev());
					msg = "  $o3 ,     .";
				}
				target->set_custom_label(label);
				act(msg, false, ch, target, nullptr, kToChar);
			}
		}
	}

	if (objname) {
		free(objname);
	}

	if (labels) {
		free(labels);
	}
}

//    :
//   ,            . 
std::string char_get_custom_label(ObjData *obj, CharData *ch) {
	const char *delim_l = nullptr;
	const char *delim_r = nullptr;

	//      
	if (obj->get_custom_label() && (ch->player_specials->clan && obj->get_custom_label()->clan_abbrev != nullptr &&
		is_alliance_by_abbr(ch, obj->get_custom_label()->clan_abbrev))) {
		delim_l = " *";
		delim_r = "*";
	} else {
		delim_l = " (";
		delim_r = ")";
	}

	std::stringstream buffer;
	if (AUTH_CUSTOM_LABEL(obj, ch)) {
		buffer << delim_l << obj->get_custom_label()->text_label << delim_r;
	}

	return buffer.str();
}

// vim: ts=4 sw=4 tw=0 noet syntax=cpp :
